/* ++++++++++
	driver.c
	Based on the Be-supplied driver.c Kernel Driver stationery
+++++ */

#include "kb_wmouse.h"

  isa_module_info *isa;
  
/* ----------
	init_hardware - called once the first time the driver is loaded
----- */
/* done */
status_t
init_hardware (void)
{
	dprintf("kb_wmouse: init_hardware()\n");
	return B_OK;
}


/* ----------
	init_driver - optional function - called every time the driver
	is loaded.
----- */
status_t
init_driver (void)
{
  int tries;
  status_t err;
  status_t result;
status_t err_kb;//-------------------
  
  dprintf("++++++++ init_driver kb_wmouse built %s %s\n", BUILD_DATE, BUILD_TIME);
  if ((err = get_module(B_ISA_MODULE_NAME, (module_info **)&isa)) != B_OK) {
    return err;
  }
  kb_buffer = cbuf_init(384L);
  if(kb_buffer == 0L) {
    put_module(B_ISA_MODULE_NAME);
    dprintf(KBWM_INIT_ERROR);
    return 0x80000000;
  }
  ms_buffer = cbuf_init(0x100);
  if(ms_buffer == 0L) {
    cbuf_delete(kb_buffer);
    put_module(B_ISA_MODULE_NAME);
    dprintf(KBWM_INIT_ERROR);
    return 0x80000000;
  }
  
  kbsem = create_sem(0, kbsem_name);  /* "kbsem" */
  if (kbsem < 0) {
    cbuf_delete(ms_buffer);
    cbuf_delete(kb_buffer);
    put_module(B_ISA_MODULE_NAME);
    dprintf(KBWM_INIT_ERROR);
    return kbsem;
  }
  
  mssem = create_sem(0, mssem_name);  /* "mssem" */
  if (mssem < 0) {
    delete_sem(kbsem);
    cbuf_delete(ms_buffer);
    cbuf_delete(kb_buffer);
    put_module(B_ISA_MODULE_NAME);
    dprintf(KBWM_INIT_ERROR);
    return mssem;
  }
  
  kb_wait_sem = create_sem(0, kb_wait_sem_name);
  if (kb_wait_sem < 0) {
    delete_sem(mssem);
    cbuf_delete(ms_buffer);
    cbuf_delete(kb_buffer);
    put_module(B_ISA_MODULE_NAME);
    dprintf(KBWM_INIT_ERROR);
    return kb_wait_sem;
  }
  
  set_sem_owner(kbsem, 2);
  set_sem_owner(mssem, 2);
  set_sem_owner(kb_wait_sem, 2);
  
  memset(key_states, 0, 32);
  memset(non_repeating, 0, 32);
  
  init_controller();
  init_keyboard();
  init_cmd_byte();
  if(disable_ps2_mouse() == 0) {
    ps2_mouse_found = 1;
    dprintf(FOUND_PS2_MOUSE);
  }
  
  control_alt_del_timeout = 0x4c4b40LL;
  reset_pending = 0;
  
  tries = 16;
  while(tries) {
    result = keyboard_command(0xF4);
    if (result == 0) break;
    dprintf(KEYB_ENABLE_RESPONSE_NOT_ACK);
    --tries;
  }
  
  dprintf(INIT_DRIVER_SUCCESS);
  return B_OK;
}


/* ----------
	uninit_driver - optional function - called every time the driver
	is unloaded
----- */
/* done */
void
uninit_driver (void)
{
  dprintf(UNINIT_MSG);
  
  if (kbsem > 0) {
    delete_sem(kbsem);
    kbsem = -1;
  }
  
  if (mssem > 0) {
    delete_sem(mssem);
    mssem = -1;
  }
  
  if (kb_wait_sem > 0) {
    delete_sem(kb_wait_sem);
    kb_wait_sem = -1;
  }
  
  if (kb_buffer != 0) {
    cbuf_delete(kb_buffer);
    kb_buffer = 0;
  }
  
  if (ms_buffer != 0) {
    cbuf_delete(ms_buffer);
    ms_buffer = 0;
  }

  put_module(B_ISA_MODULE_NAME);
  
  return;
}

/* kb_wmouse_open() */
static status_t kb_wmouse_open(const char *name,
                              uint32 flags,
                              void **cookie) {
  
  dprintf(KB_MOUSE_OPEN, (name != 0) ? name : "NULL");
  if(strcmp(name, MOUSE_SERIAL_NAME_FULL) == 0)  {
    *(int32 *)cookie = 'seri';
    init_serial_mouse();
    return B_OK;
  }
  if(strcmp(name, PS2_MOUSE_NAME_FULL) == 0) {
    if(ps2_mouse_found == 0) return ENODEV;
    *(int32 *)cookie = 'mous';
    enable_ps2_mouse();
    install_io_interrupt_handler(12L, ms_inth, 0L, 0L);
    change_command_bits(2, 2);
    return B_OK;
  }
  
  if(strcmp(name, KB_NAME_FULL) == 0) {
    if(found_keyboard == 0) return ENODEV;
    else {
      dprintf(KB_OPENING_AT_KEYBOARD);
      *(int32 *)cookie = 'kbd ';
      install_io_interrupt_handler(1L, kb_inth, 0L, 0L);
      change_command_bits(0x45, 0x45);
      set_key_repeat_info(typematic_rate, repeat_delay);
      return B_OK;
    }
  }
  return ENODEV;
}  /* kb_wmouse_open() */

/* kb_wmouse_read() */
static status_t kb_wmouse_read(void *cookie,
                               off_t position,
                               void *buf,
                               size_t *num_bytes) {
  *num_bytes = 0L;
  return B_BAD_VALUE;
}  /* kb_wmouse_read() */

/* kb_wmouse_write() */
static status_t kb_wmouse_write(void *cookie,
                                off_t position,
                                const void *buffer,
                                size_t *num_bytes) {
  *num_bytes = 0L;
  return B_BAD_VALUE;
}  /* kb_wmouse_write() */

/* kb_wmouse_control() */
static status_t kb_wmouse_control(void *cookie,
                                 uint32 which_ioctl,
                                 void *arg,
                                 size_t len) {
  status_t err;
  int32 mssem_count;
  
  switch(which_ioctl) {
    case 9999:
      err = acquire_sem_etc(kbsem, 1, B_CAN_INTERRUPT, 0LL);
      if(err != B_OK) return err;
      cbuf_getn(kb_buffer, k_locked_66, 12);
      memcpy(arg, k_locked_66, 12);
      return err;
      break;
    case 10000:
      return active_keyboard_command_get_results(0xF2, arg, 2);
      break;
    case 10001:  /* set keyboard LEDs */
      return active_keyboard_command_with_data(0xED,
          (((char *)arg)[0] ? 2 : 0)
        | (((char *)arg)[1] ? 4 : 0)
        | (((char *)arg)[2] ? 1 : 0));
      break;
    case 10002:  /* turn off repeating for key *arg */
      if((*(int32 *)arg > 0x100) || (*(int32 *)arg < 0)) return 0x80000005;
      non_repeating[(*(int32 *)arg) >> 3] &= !(1 << ((7 - *(int32 *)arg) & 7));
      return B_OK;
      break;
    case 10003:
      if((*(int32 *)arg > 0x100) || (*(int32 *)arg < 0)) return 0x80000005;
      non_repeating[(*(int32 *)arg) >> 3] |= 1 << ((7 - *(int32 *)arg) & 7);
      return B_OK;
      break;
    case 10004:
dprintf("kb_wmouse: set_key_repeat_info(%ld, -1LL)\n", *(int32 *)arg);
      set_key_repeat_info(*(int32 *)arg, -1LL);
      return B_OK;
      break;
    case 10005:
dprintf("kb_wmouse: get_typematic_rate()\n");
      *(int32 *)arg = typematic_rate;
      return B_OK;
      break;
    case 10006:
dprintf("kb_wmouse: set_key_repeat_info(-1, %Ld\n", *(int64 *)arg);
      set_key_repeat_info(-1, *(int64 *)arg);
      return B_OK;
      break;
    case 10007:
dprintf("kb_wmouse: get_repeat_delay()\n");
      *(int64 *)arg = repeat_delay;
      return B_OK;
      break;
    case 10008:
      control_alt_del_timeout = *(int64 *)arg;
      return B_OK;
      break;
    case 10009:
      cancel_timer(&reset_timer);
      return B_OK;
      break;
    case 10010:
      cancel_timer(&reset_timer);
      reset_pending = 0;
      return B_OK;
      break;
    case 10099:
//      if((((struct mouse_movement *)arg)->ser_fd_index == -1) || (ser_fd_list == 0L)) err = do_ps2_mouse_read();
      if((((struct mouse_movement *)arg)->ser_fd_index == -1) || (ser_fd_list == 0L)) {
err = do_ps2_mouse_read();
}
//      else err = do_ser_mouse_read(ser_fd_list[((struct mouse_movement *)arg)->ser_fd_index]);
      else {
dprintf("kb_wmouse: do_ser_mouse_read(0x%08x; index %d)\n", arg, ((struct mouse_movement *)arg)->ser_fd_index);
err = do_ser_mouse_read(ser_fd_list[((struct mouse_movement *)arg)->ser_fd_index]);
}
//      if(err != B_OK) return err;
      if(err != B_OK) {
dprintf("kb_wmouse: error 0x%x from do_foo_mouse_read()\n");
return err;
}
      ((struct mouse_movement *)arg)->xdelta = mouse_xdelta;
      ((struct mouse_movement *)arg)->ydelta = mouse_ydelta;
      ((struct mouse_movement *)arg)->buttons = 0L;
      if((ms_type == 1) && (mouse_buttons != 0L)) {
        mouse_buttons = 1L;
      } else if((ms_type == 2) && !(mouse_buttons & MMB)) {  /* middle physical button */
        if(mouse_mapping[0] != 2) {
          mouse_buttons = (mouse_buttons & LMB) | RMB;
        } else {
          mouse_buttons = (mouse_buttons & LMB) | RMB;
        }
      }
      if(mouse_buttons & LMB) ((struct mouse_movement *)arg)->buttons |= mouse_mapping[0];
      if(mouse_buttons & RMB) ((struct mouse_movement *)arg)->buttons |= mouse_mapping[1];
      if(mouse_buttons & MMB) ((struct mouse_movement *)arg)->buttons |= mouse_mapping[2];
      ((struct mouse_movement *)arg)->click_count = click_count;
      ((struct mouse_movement *)arg)->mouse_mods = mouse_mods;
      ((struct mouse_movement *)arg)->mouse_time = mouse_time;
      return B_OK;
      break;
    case 10100:
      if(((int32)arg == -1) || (ser_fd_list == 0L)) {
        get_sem_count(mssem, &mssem_count);
        return mssem_count;
      } else {
dprintf("kb_wmouse: TCWAITEVENT\n");
        return ioctl(ser_fd_list[(int32)arg], TCWAITEVENT);
      }
      break;
    case 10101:
dprintf("kb_wmouse: get acceleration info\n");
      ((int32 *)arg)[0] = accel[0];
      ((int32 *)arg)[1] = accel[1];
      ((int32 *)arg)[2] = accel[0];
      return B_OK;
      break;
    case 10102:
dprintf("kb_wmouse: set acceleration info\n");
      accel[0] = ((int32 *)arg)[0];
      accel[1] = ((int32 *)arg)[1];
      accel[2] = ((int32 *)arg)[2];
      return B_OK;
      break;
    case 10103:
dprintf("kb_wmouse: get_mouse_type()\n");
      *(int32 *)arg = ms_type;
      return B_OK;
      break;
    case 10104:
dprintf("kb_wmouse: set_mouse_type()\n");
      if(*(int32 *)arg <= 3) {
        ms_type = *(int32 *)arg;
        return B_OK;
      } else {
        return B_ERROR;
      }
      break;
    case 10105:
dprintf("kb_wmouse: get_mouse_mapping()\n");
      ((int32 *)arg)[0] = mouse_mapping[0];
      ((int32 *)arg)[1] = mouse_mapping[1];
      ((int32 *)arg)[2] = mouse_mapping[2];
      return B_OK;
      break;
    case 10106:
dprintf("kb_wmouse: set_mouse_mapping\n");
      if(((int32 *)arg)[0] & 7) {
        if(((int32 *)arg)[0] != 3)
          mouse_mapping[0] = ((int32 *)arg)[0];
      }
      if(((int32 *)arg)[1] & 7) {
        if(((int32 *)arg)[1] != 3)
          mouse_mapping[1] = ((int32 *)arg)[1];
      }
      if(((int32 *)arg)[2] & 7) {
        if(((int32 *)arg)[2] != 3)
          mouse_mapping[2] = ((int32 *)arg)[2];
      }
      return B_OK;
      break;
    case 10107:
dprintf("kb_wmouse: get_click_speed()\n");
      *(int64 *)arg = click_speed;
      return B_OK;
      break;
    case 10108:
dprintf("kb_wmouse: set_click_speed()\n");
      click_speed = *(int64 *)arg;
      return B_OK;
      break;
    case 10109:
      return ser_fd_count;
      break;
    default:  /* bad_ioctl */
dprintf("kb_wmouse: unrecognised ioctl %d\n", which_ioctl);
      return B_ERROR;
  }
}  /* kb_wmouse_control() */

/* kb_wmouse_close() */
static status_t kb_wmouse_close(void *cookie) {
  int32 type = (int32)cookie;
  int i = 0;
  
  dprintf(KB_WMOUSE_CLOSE);
  if(type == 'seri') {
    dprintf(SER_MOUSE_CLOSING);
    if(i < ser_fd_count) {
      while(i < ser_fd_count) {
        close(ser_fd_list[i]);
        ++i;
      }
      ser_fd_count = 0;
      free(ser_fd_list);
      ser_fd_list = 0L;
    }
  }
  return B_OK;
}  /* kb_wmouse_close() */

/* kb_wmouse_free() */
static status_t kb_wmouse_free(void *cookie) {
  int32 type = (int32)cookie;
  
  dprintf(KB_WMOUSE_FREE);
  switch(type) {
    case 'mous':
      change_command_bits(2, 0);
      disable_ps2_mouse();
      remove_io_interrupt_handler(12, ms_inth, 0);
      return B_ERROR;
      break;
    case 'kbd ':
      change_command_bits(1, 0);
      remove_io_interrupt_handler(1, kb_inth, 0);
      return B_ERROR;
      break;
    case 'seri':
      return B_OK;
      break;
    default: return B_ERROR;
  }
}  /* kb_wmouse_free() */


/* -----
	function pointers for the device hooks entry points
----- */
device_hooks kbm_device = {
	kb_wmouse_open, 			/* -> open entry point */
	kb_wmouse_close, 			/* -> close entry point */
	kb_wmouse_free,			/* -> free cookie */
	kb_wmouse_control, 		/* -> control entry point */
	kb_wmouse_read,			/* -> read entry point */
	kb_wmouse_write			/* -> write entry point */
};

/* publish_devices() */
const char **publish_devices(void) {
  int i = 0;
  
  dprintf("kb_wmouse: publish_devices()\n");
dprintf("kb_wmouse: found_keyboard == %d\n", found_keyboard);
  if(found_keyboard != 0) {
    ++i;
    kb_device_name[0] = (char *)KB_NAME_FULL;
  }
  if(ps2_mouse_found != 0) {
    kb_device_name[i] = PS2_MOUSE_NAME_FULL;
    ++i;
  }
  kb_device_name[i] = MOUSE_SERIAL_NAME_FULL;
  kb_device_name[i+1] = 0L;
  
dprintf("kb_wmouse: publishing %d devices\n", i + 1);
  return (const char **)kb_device_name;
}  /* publish_devices() */

/* find_device() */
device_hooks *find_device(const char *name) {
  return &kbm_device;
}  /* find_device() */
