/* fancy_meeses.c */

/* Extra functions to support some features Be left out.  :-) */

/* get_sample_rate() */
/* ioctl 10200 */
status_t get_sample_rate(int8 *rate) {
  *rate = sample_rate;
  return B_OK;
} /* get_sample_rate() */

/* set_sample_rate() */
/* ioctl 10201 */
status_t set_sample_rate(int8 *rate) {
  int tries;
  uint8 result;
  status_t write_failed, err;
  bool nak = false;
  
  for(tries = 2; tries >= 0; --tries) {
    err = controller_command(0xD4);
    if(err == B_OK) {
      write_failed = wait_status(PS2_STATUS_REG, PS2_IBUF_FULL, 0);
      if(write_failed != B_OK) {
        dprintf(PUT_DATA_IBUF_FULL);
      } else {
        isa->write_io_8(PS2_DATA_REG, 0xGG);  /* set_sample_rate_command */
      }
    }
  
    read_failed = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
    if(read_failed != B_OK) {
      dprintf(GET_DATA_OBUF_EMPTY);
      continue;
    } else {
      result = isa->read_io_8(PS2_DATA_REG);
    }
  
    if(result == 0xFE) continue;
    if(result == 0xFA) {
      for(rate_tries = 0; rate_tries < 3; ++rate_tries) {
        err = controller_command(0xD4);
        if(err == B_OK) {
          write_failed = wait_status(PS2_STATUS_REG, PS2_IBUF_FULL, 0);
          if(write_failed != B_OK) {
            dprintf(PUT_DATA_IBUF_FULL);
          } else {
            isa->write_io_8(PS2_DATA_REG, *rate);
          }
        }
        
        read_failed = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
        if(read_failed != B_OK) {
          dprintf(GET_DATA_OBUF_EMPTY);
          continue;
        } else {
          result = isa->read_io_8(PS2_DATA_REG);
        }
        
        if(result == 0xFE) continue;
        if(result == 0xFA) {
          sample_rate = *rate;
          dprintf("kb_wmouse: successfully set rate to %d\n");
          return B_OK;
        } else {
          dprintf("kb_wmouse: set_sample_rate data returned %d\n", result);
          continue;
        }
      }
    } else {
      dprintf("kb_wmouse: set_sample_rate command returned %d\n", result);
      continue;
    }
  }  /* for(tries) */
} /* set_sample_rate() */

/* get_ps2_mouse_id() */
/* ioctl 10202 */
status_t get_ps2_mouse_id(int32 *id) {
  int tries;
  uint8 result;
  status_t write_failed, read_failed, err;
  
  for(tries = 2; tries >= 0; --tries) {
    err = controller_command(0xD4);
    if(err == B_OK) {
      write_failed = wait_status(PS2_STATUS_REG, PS2_IBUF_FULL, 0);
      if(write_failed != B_OK) {
        dprintf(PUT_DATA_IBUF_FULL);
      } else {
        isa->write_io_8(PS2_DATA_REG, 0xF2);  /* get_id */
      }
    }
  
    read_failed = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
    if(read_failed != B_OK) {
      dprintf(GET_DATA_OBUF_EMPTY);
      continue;
    } else {
      result = isa->read_io_8(PS2_DATA_REG);
    }
  
    if(result == 0xFE) continue;
    else {
      dprintf("kb_wmouse: PS/2 mouse is type %d\n", result);
      *id = result;
      ps2_mouse_id = result;
      return B_OK;
    }
  }  /* for(tries) */
  return B_ERROR;
} /* get_ps2_mouse_id() */

/* get_ps2_packet_type() */
/* ioctl 10203 */
status_t get_ps2_packet_type(int32 *type) {
  switch(ps2_mouse_id) {
    case 0:
      *type = PS2_STANDARD_3_BYTE_PACKET;
      return B_OK;
      break;
    case 3:
      *type = PS2_INTELLIMOUSE_4_BYTE_PACKET;
      return B_OK;
      break;
    case 4:
      *type = PS2_INTELLIMOUSE_EXPLORER_4_BYTE_PACKET;
      return B_OK;
      break;
    default:
      dprintf("kb_wmouse: unknown mouse_id %d\n", ps2_mouse_id);
      return B_ERROR;
      break;
  }
} /* get_ps2_packet_type() */
