/* guts.c                               */
/* This is where the fun stuff happens. */

#include "kb_wmouse.h"

/* do_ps2_mouse_read() */
status_t do_ps2_mouse_read(void) {
  status_t err;
  int bytes_read;
  int32 x_delta, y_delta;
  
  err = acquire_sem_etc(mssem, 1, B_CAN_INTERRUPT, 0LL);
//dprintf("kb_wmouse:do_ps2_mouse_read(): err == 0x%lx\n", err);
  if(err != B_OK) return err;
  
  bytes_read = cbuf_getn(ms_buffer, buff_46, 3);
  if(bytes_read != 3) {
    dprintf(PS2_YIKES_NOT_3_MOUSE_BYTES);
    return B_ERROR;
  }
  
  x_delta = ((buff_46[0] & 0x10) ? 0xFFFFFF00 : 0) | buff_46[1];
  y_delta = ((buff_46[0] & 0x20) ? 0xFFFFFF00 : 0) | buff_46[2];
  accel_and_stash(x_delta, y_delta, buff_46[0] & 7);
  return B_OK;
}  /* do_ps2_mouse_read() */

status_t do_ser_mouse_read(int ser_fd) {
  int tries;
  int bytes_read;
  int32 x_delta, y_delta, buttons;

  for(tries = 0; tries <= 5; ++tries) {
    bytes_read = read(ser_fd, buffer_49, 3);
    if(bytes_read < 0) {
      dprintf(MOUSE_BAD_READ_SER);
      return B_ERROR;
    }
    if(!(buffer_49[0] & 0x40)) {
      dprintf(MOUSE_YIKES_NOT_SER_MOTION, buffer_49[0], buffer_49[1], buffer_49[2]);
      if(buffer_49[1] & 0x40) {
        buffer_49[0] = buffer_49[1];
        buffer_49[1] = buffer_49[2];
        read(ser_fd, &(buffer_49[2]), 1);
      } else {
        if(buffer_49[2] & 0x40) {
          buffer_49[0] = buffer_49[2];
          read(ser_fd, &(buffer_49[1]), 2);
        } else {
          continue;
        }
      }
    }
  
    if((buffer_49[1] & 0x40) || (buffer_49[2] & 0x40)) {
      dprintf(MOUSE_BAD_BYTES, buffer_49[1], buffer_49[2]);
    }
  
    x_delta = buffer_49[0] << 6 | (buffer_49[1] & 0x3F);
    y_delta = ((buffer_49[0] & 0x0C) << 4) | (buffer_49[2] & 0x3F);
/*    buttons = (buffer_49[0] >> 3) & 0x02;*/  /* bit 4 */
/*    if(buffer_49[0] & 0x20) buttons |= 1;  *//* bit 5 */
    buttons = ((buffer_49[0] & 0x10) ? 0x02 : 0)
            | ((buffer_49[0] & 0x20) ? 0x01 : 0);
    accel_and_stash((int32)x_delta, -(int32)y_delta, buttons);  /* both sign extended */
    return B_OK;
  }
  return B_ERROR;  
}  /* do_ser_mouse_read() */

/* set_key_repeat_info() */
status_t set_key_repeat_info(int32 typematic, int64 delay) {
  uint8 send_value = 0x7F;
  int i;
  
  if(typematic < 0) typematic = typematic_rate;
  if(delay < 0) delay = repeat_delay;

  for(i = 0; i <= 0x1F; ++i) {
    if(valid_rates[i] <= typematic) {
      typematic_rate = typematic;
      send_value = i;
      break;
    }
  }
  
  for(i = 0; i <= 3; ++i) {
    if(delay <= valid_rates[i]) {
      repeat_delay = delay;
      send_value = (send_value & 0x1F) | (i << 5);
      break;
    }
  }
  
dprintf("kb_wmouse:set_key_repeat_info(): value is %d\n", send_value);
  active_keyboard_command_with_data(0xF3, send_value);
  return B_OK;
}

/* kb_inth() */
long kb_inth(void *data) {
  uint8 value_read, value_bit7, corrected_value, bit_number;
  status_t err;
  int32 old_waiting;
  struct kb_inth_struct stuff_to_put_to_buffer;

  value_read = isa->read_io_8(PS2_STATUS_REG);
  if((int8)value_read < 0) {
    dprintf(KB_INTH_DET_PROBLEM, value_read);
    unhandled_interrupt_hook_8042();
    return B_HANDLED_INTERRUPT;
  }
  err = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
  if(err != B_OK) {
    dprintf(GET_DATA_OBUF_EMPTY);
  } else {
    value_read = isa->read_io_8(PS2_DATA_REG);
  }
  
  old_waiting = atomic_add(&waiting, -1);
  if(old_waiting > 0) {
    keyboardResult[old_waiting - 1] = value_read;
    if(old_waiting != 1) return B_HANDLED_INTERRUPT;
    release_sem_etc(kb_wait_sem, 1, B_DO_NOT_RESCHEDULE);
    return B_INVOKE_SCHEDULER;
  }
  if(value_read == 0xE1) return B_HANDLED_INTERRUPT;
  if(value_read == 0xE0) {
    extended_key_75 = 1;
    return B_HANDLED_INTERRUPT;
  }
  
  value_bit7 = (value_read & 0x80) ? 0 : 1;
  corrected_value = value_read & 0x7F;
  if(extended_key_75 != 0) {
    corrected_value -= 0x80;
  }
  value_read = corrected_value;
  extended_key_75 = 0;
  if(value_read == 0x54) {
    if(value_bit7 != 0) {
      kernel_debugger(SYSTEM_REQUEST);
      return B_HANDLED_INTERRUPT;
    }
    return B_HANDLED_INTERRUPT;
  }
  
  if(value_bit7 != 0) {
    if((value_read == 0xF1) || (value_read == 0xD3)) {
      if((key_states[3] & 4) && ((int8)key_states[7] < 0)) {
        if(reset_pending != 0) {
          cancel_timer(&reset_timer);
          do_control_alt_del();
        }
        add_timer(&reset_timer, reset_timer_handler, control_alt_del_timeout, 2);
        reset_pending = 1;
      }
    }
    if(value_bit7 != 0) {
      bit_number = (7 - value_read) & 7;
      if(key_states[value_read >> 3] & (1 << bit_number)) {
        if(!(non_repeating[value_read >> 3] & (1 << bit_number))) {
          goto kb_inth_end;
        } else {
          return B_HANDLED_INTERRUPT;
        }
      }
      key_states[value_read >> 3] |= (1 << bit_number);
    }
  } else {  /* value_bit7 == 1 => key_up */
    key_states[value_read >> 3] &= !(1 << ((7 - value_read) & 7));
  }
  kb_inth_end:
  stuff_to_put_to_buffer.current_time = system_time();
  stuff_to_put_to_buffer.value_read = value_read;
  stuff_to_put_to_buffer.value_bit7 = value_bit7;
  cbuf_putn(kb_buffer, (char *)&stuff_to_put_to_buffer, sizeof(struct kb_inth_struct));
  release_sem_etc(kbsem, 1, B_DO_NOT_RESCHEDULE);
  return B_INVOKE_SCHEDULER;
}  /* kb_inth() */

/* ms_inth() */
long ms_inth(void *data) {
  int8 read_byte;
  status_t read_failed;
  
  read_byte = isa->read_io_8(PS2_STATUS_REG);
  if(read_byte < 0) {
    dprintf(MOUSE_INTH_DET_PROBLEM, read_byte);
    unhandled_interrupt_hook_8042();
    return B_HANDLED_INTERRUPT;
  }
  
  read_failed = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
  if(read_failed != B_OK) {
    dprintf(GET_DATA_OBUF_EMPTY);
  } else {
    read_byte = isa->read_io_8(PS2_DATA_REG);
  }
  
  if((mouse_sync == 0) && !(read_byte & 8)) {
    dprintf(MOUSE_RESYNCED_BAD_DATA, read_byte);
    return B_HANDLED_INTERRUPT;
  }
  
  cbuf_put(ms_buffer, read_byte);
  if(mouse_sync++ == 2) {
dprintf(" %02x\n", read_byte & 0xff);
    mouse_sync = 0;
    release_sem_etc(mssem, 1, B_DO_NOT_RESCHEDULE);
    return B_INVOKE_SCHEDULER;
  }
else
dprintf(" %02x", read_byte & 0xff);
  return B_HANDLED_INTERRUPT;
}  /* ms_inth() */

/* accel_and_stash() */
/* Now with no acceleration! */
void accel_and_stash(int32 x_delta, int32 y_delta, int32 buttons) {
  int64 current_time;
  
dprintf("as(%ld, %ld, %ld)\n", x_delta, y_delta, buttons);
  
  current_time = system_time();
  if(buttons != 0) {
    if(mouse_buttons == 0) {
      if((last_click + click_speed) > current_time) {
        ++click_count;
      } else {
        click_count = 1;
      }
    }
    last_click = current_time;
  }
  mouse_buttons = buttons;
  mouse_xdelta = x_delta;
  mouse_ydelta = y_delta;
  mouse_mods = 0;
  mouse_time = current_time;
  return;
}  /* accel_and_stash() */

/* unhandled_interrupt_hook_8042() */
void unhandled_interrupt_hook_8042(void) {
  uint8 status, status_again, byte_read, io_err, controller_mode;
  status_t err, read_failed, write_failed;
  
  status = isa->read_io_8(PS2_STATUS_REG);
  dprintf(KB_UHI_STATUS, status);
  if(!(status & 0xC0)) {
    dprintf(KB_NO_PARITY_OR_TIMEOUT_ERROR);
    if(status & 0x20) {
      if(status & 0x01) {
        ms_inth(0);
        return;
      }
    } else {
      if(status & 0x01) {
        kb_inth(0);
        return;
      }
    }
    isa->write_io_8(0x398, 5);
    io_err = isa->read_io_8(0x399);
    dprintf(KRR, io_err);
    err = controller_command(0x20);  /* KBD_CCMD_READ_MODE */
    if(err != B_OK) {
      status = isa->read_io_8(PS2_STATUS_REG);
      dprintf(KBC_RDCOMMAND_FAILED_STATUS, status);
      return;
    }
    read_failed = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
    if(read_failed != B_OK) {
      dprintf(GET_DATA_OBUF_EMPTY);
    } else {
      controller_mode = isa->read_io_8(PS2_DATA_REG);
    }
    if(read_failed) {
      status = isa->read_io_8(PS2_STATUS_REG);
      dprintf(GET_DATA_FOR_KBC_RD_CMD_FAILED, status);
      return;
    } else {
      dprintf(COMMAND_BYTE, controller_mode);
      kernel_debugger(0);
      return;
    }
  }

  /* Right, we know we've got a timeout or parity error now. */
  do {
    status_again = isa->read_io_8(PS2_STATUS_REG);
    if(status_again & PS2_MOUSE_OBUF_FULL) {
      if(status_again & PS2_OBUF_FULL) {
        read_failed = wait_status(PS2_STATUS_REG, PS2_MOUSE_OBUF_FULL, PS2_MOUSE_OBUF_FULL);
        if(read_failed != B_OK) {
          dprintf(GET_DATA_OBUF_EMPTY);
        } else {
          byte_read = isa->read_io_8(PS2_DATA_REG);
        }
        dprintf(ODS_msg, byte_read);
      } else break;  /* !PS2_OBUF_FULL */
    } else {  /* !PS2_MOUSE_OBUF_FULL */
      if(status_again & PS2_OBUF_FULL) {
        read_failed = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
        if(read_failed != B_OK) {
          dprintf(GET_DATA_OBUF_EMPTY);
        } else {
          byte_read = isa->read_io_8(PS2_DATA_REG);
        }
        dprintf(OBF_msg, byte_read);
      } else break;  /* !PS2_OBUF_FULL */
    }
  } while(status_again & PS2_OBUF_FULL);
  if(controller_command(0x20) != B_OK) {
    dprintf(GET_DATA_FOR_KBC_RD_CMD_FAILED);
    return;
  }
  
  read_failed = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
  if(read_failed != B_OK) {
    dprintf(GET_DATA_OBUF_EMPTY);
    dprintf(GET_DATA_FOR_KBC_RD_CMD_FAILED);
    return;
  } else {
    byte_read = isa->read_io_8(PS2_DATA_REG);
  }
  
  dprintf(COMMAND_BYTE, byte_read);
  if(!(status & PS2_MOUSE_OBUF_FULL)) {
    if(status & PS2_OBUF_FULL) return;
    if(keyboard_command(0xFE) != B_OK) {
      dprintf(RESEND_KB_FAILED);
      return;
    } else {
      dprintf(RESEND_KB_OK);
      return;
    }
  } else {  /* PS2_MOUSE_OBUF_FULL */
    if(!(status & PS2_OBUF_FULL)) return;
    do {
      if(controller_command(0xD4) != B_OK) {
        write_failed = wait_status(PS2_STATUS_REG, PS2_IBUF_FULL, 0);
        if(write_failed != B_OK) {
          dprintf(PUT_DATA_IBUF_FULL);
        } else {
          isa->write_io_8(PS2_DATA_REG, 0xFE);
        }
      }
      read_failed = wait_status(PS2_STATUS_REG, PS2_OBUF_FULL, PS2_OBUF_FULL);
      if(read_failed != B_OK) {
        dprintf(GET_DATA_OBUF_EMPTY);
        dprintf(RESEND_MS_FAILED);
        return;
      } else {
        status_again = isa->read_io_8(PS2_DATA_REG);
      }
      if(status_again == 0xFE) {
        dprintf(RESEND_RESEND_MS);
        continue;
      }
      if(status_again == 0xFA) {
        dprintf(RESEND_MS_OK);
        return;
      }
    } while(status_again == 0xFE);
  }
}  /* unhandled_interrupt_hook_8042() */

/* ps2_mouse_num_events() */
int32 ps2_mouse_num_events(void) {
  int32 mssem_count;
  get_sem_count(mssem, &mssem_count);
  return mssem_count;
}  /* ps2_mouse_num_events() */

/* ser_mouse_num_events() */
int32 ser_mouse_num_events(int32 ser_fd) {
  char dummy;
  return ioctl(ser_fd, TCWAITEVENT, &dummy);
}  /* ser_mouse_num_events() */

/* reset_timer_handler */
int32 reset_timer_handler(timer *some_timer) {
  do_control_alt_del();
  return 1;
}  /* reset_timer_handler() */
