#ifndef KB_WMOUSE_H
#define KB_WMOUSE_H

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <drivers/ISA.h>
#include <termios.h>
#include <dirent.h>
#include <fcntl.h>

#define LMB (1)
#define RMB (2)
#define MMB (4)

struct mouse_movement {
  int32 ser_fd_index;
  int32 buttons;
  int32 xdelta;
  int32 ydelta;
  int32 click_count;
  int32 mouse_mods;
  int64 mouse_time;
};

#define PS2_STATUS_REG (0x64)
#define PS2_CMD_REG    (0x64)
#define PS2_DATA_REG   (0x60)

#define PS2_OBUF_FULL (1)
#define PS2_IBUF_FULL (2)
#define PS2_MOUSE_OBUF_FULL (0x20)

#define PS2_MAX_IOCTL (10109)

struct kb_inth_struct {
  int64 current_time;
  int8 value_read;
  int8 value_bit7;
};

char *kb_device_name[4];

/* Undocumented kernel cbuf_* functions */
struct cbuffer_t;
typedef struct cbuffer_t cbuffer;

size_t cbuf_getn_no_lock(cbuffer *, char *, size_t);
size_t cbuf_putn_no_lock(cbuffer *, char *, size_t);
cbuffer *cbuf_init(size_t size);
void cbuf_delete(cbuffer *buffer);
char cbuf_get(cbuffer *);
bool cbuf_mt(cbuffer *);
bool cbuf_full(cbuffer *);
status_t cbuf_put(cbuffer *, char);
status_t cbuf_unput(cbuffer *);
void cbuf_flush(cbuffer *);
size_t cbuf_size(cbuffer *);
size_t cbuf_avail(cbuffer *);
size_t cbuf_free(cbuffer *);
size_t cbuf_putn(cbuffer *, char *, size_t num_bytes);
size_t cbuf_getn(cbuffer *, char *, size_t num_bytes);
cpu_status cbuf_lock(cbuffer *);
void cbuf_unlock(cbuffer *, cpu_status);



/* prototypes */

static status_t kb_wmouse_open(const char *name,
                              uint32 type,
                              void **cookie);
static status_t kb_wmouse_read(void *cookie,
                               off_t position,
                               void *buf,
                               size_t *num_bytes);
static status_t kb_wmouse_write(void *cookie,
                                off_t position,
                                const void *buffer,
                                size_t *num_bytes);
static status_t kb_wmouse_control(void *cookie,
                                 uint32 which_ioctl,
                                 void *arg,
                                 size_t len);
static status_t kb_wmouse_close(void *cookie);
static status_t kb_wmouse_free(void *cookie);

_EXPORT status_t init_hardware(void);
_EXPORT const char **publish_devices();
_EXPORT device_hooks *find_device(const char *name);
_EXPORT status_t init_driver(void);
_EXPORT void uninit_driver(void);

status_t do_ps2_mouse_read(void);
status_t do_ser_mouse_read(int ser_fd);
status_t set_key_repeat_info(int32 typematic, int64 delay);
void accel_and_stash(int32 second, int32 third, int32 first);
void unhandled_interrupt_hook_8042(void);
int32 ps2_mouse_num_events(void);
int32 ser_mouse_num_events(int32 ser_fd);
int32 reset_timer_handler(timer *);
long ms_inth(void *data);
long kb_inth(void *data);

/* strings */

#define GET_DATA_OBUF_EMPTY ("get_data error: output buffer empty!\n")
#define FLUSHED_MSG ("flushed %.2x\n")
#define PUT_DATA_IBUF_FULL ("put_data error: input buffer not empty!\n")
#define CONTROLLER_COMMAND_IBUF_NOT_EMPTY ("controller_command: input buffer not empty!\n")
#define KB_CMD_ERROR ("!!!! keyboard_command error! %x %x\n")
#define KB_CMD_WITH_DATA_ERROR ("!!!! keyboard_command_with_data error! %x %x\n")
#define KB_CNTRL_TEST_FAILED_2X ("kb controller test failed: %.2x\n")
#define KB_CNTRL_TEST_FAILED ("kb controller test failed\n")
#define ERROR_ENABLING_KB ("error enabling kb\n")
#define ERROR_ENABLING_MOUSE ("error enabling mouse\n")
#define RESP_KB_DIS_NOT_ACK ("Response to keyboard disable was not ACK!\n")
#define KEYB_ENABLE_RESPONSE_NOT_ACK ("Response to keyboard enable was not ACK!\n")
#define RESP_MOUSE_DIS_NOT_ACK ("Response to mouse disable was not ACK! (was %.2x)\n")
#define FOUND_PS2_MOUSE ("Found a PS\\2 mouse!\n")
//#define DEV_PORTS_DIR ("/dev/ports")
#define aDevPorts_0 ("/dev/ports/")
//#define a_ (".")
//#define a__ ("..")
#define KB_WMOUSE_OPENING ("kb_wmouse: opening %s\n")
#define KB_WMOUSE_BUSY ("kb_wmouse: %s busy\n")
#define KB_WMOUSE_SERIAL_MOUSE_ON ("kb_wmouse: Serial mouse on %s\n")
#define KB_MOUSE_NO_MOUSE_AT ("kb_wmouse: No mouse at %s\n")
#define PS2_SUCC_ENABLE_MOUSE ("SUccessfully enabled PS\\2 mouse\n")
#define PS2_BAD_ENABLE_RESPONSE ("Response to mouse enable was not ACK nor resend! (was %.2x)\n")
#define PS2_YIKES_NOT_3_MOUSE_BYTES ("Yikes! didn't read 3 bytes from ms_buffer\n")
#define MOUSE_BAD_READ_SER ("mouse: bad read from serial device!\n")
#define MOUSE_YIKES_NOT_SER_MOTION ("Yikes! not ser-mouse motion data (0x%x 0x%x 0x%x)!\n")
#define MOUSE_BAD_BYTES ("BAD: buffer[1] 0x%x buffer[2] 0x%x\n")
#define KB_WMOUSE_CLOSE ("kbwmouse close\n")
#define SER_MOUSE_CLOSING ("closing mouse_serial!\n")
#define KB_WMOUSE_FREE ("kbwmouse free\n")
#define BUILD_TIME ("14:10:47")
#define BUILD_DATE ("Aug 24 1999")
#define aInit_driverKb ("++++++++ init_driver kb_mouse built %s %s\n")
#define aKbsem ("kbsem")
#define aMssem ("mssem")
#define aKb_wait_sem ("kb_wait_sem")
#define INIT_DRIVER_SUCCESS ("kbwmouse init_driver: success\n")
#define KBWM_INIT_ERROR ("kb_wmouse initialisation error\n")
#define UNINIT_MSG ("uninit_driver kb_wmouse\n")
#define aNULL ("NULL")
#define KB_MOUSE_OPEN ("kbwmouse open '%s'\n")
#define KB_OPENING_AT_KEYBOARD ("opening AT keyboard\n")
#define KB_INTH_DET_PROBLEM ("keyboard inth detected a problem, status: %.2x\n")
#define SYSTEM_REQUEST ("System request\n")
#define MOUSE_INTH_DET_PROBLEM ("mouse inth detected a problem, status: %.2x\n")
#define MOUSE_RESYNCED_BAD_DATA ("Mouse re-synced (bad data -> %x)\n")
#define KB_UHI_STATUS ("kb-uhi: status=%.2x\n")
#define KB_NO_PARITY_OR_TIMEOUT_ERROR ("        no parity or timeout error!\n")
#define KRR ("        KRR: %.2x\n")
#define KBC_RDCOMMAND_FAILED_STATUS ("        KBC_RDCOMMAND failed-status = %.2x\n")
#define GET_DATA_FOR_KBC_RD_CMD_FAILED ("        get_data for KBC_RD_COMMAND failed-status = %.2x\n")
#define COMMAND_BYTE ("        command byte=%.2x\n")
#define ODS_msg ("        ODS: %x\n")
#define OBF_msg ("        OBF: %x\n")
#define KBC_RDCMD_faile ("        KBC_RDCOMMAND failed\n")
#define RESEND_MS_FAILED ("        resend command to mouse failed\n")
#define RESEND_RESEND_MS ("      resend `resend'' command to mouse\n")
#define RESEND_MS_OK ("        resend command to mouse OK\n")
#define RESEND_KB_OK ("        resend command to keyboard OK\n")
#define RESEND_KB_FAILED ("        resend command to keyboard failed\n")
#define KB_NAME_FULL (kATKeyboardDeviceName)
#define PS2_MOUSE_NAME_FULL (kPS2MouseDeviceName)
#define MOUSE_SERIAL_NAME_FULL (kSerialMouseDeviceName)

/* globals */

extern char kATKeyboardDeviceName[];
extern char kPS2MouseDeviceName[];
extern char kSerialMouseDeviceName[];
extern char isa_name[];
extern cbuffer *kb_buffer;
extern cbuffer *ms_buffer;
extern int32 waiting;
extern sem_id kb_wait_sem;
extern int32 mouse_buttons;
extern int32 mouse_xdelta;
extern int32 mouse_ydelta;
extern int32 click_count;
extern int64 last_click;
extern int32 mouse_mods;
extern int32 mouse_sync;
extern int64 mouse_time;
extern int16 mouse_plugged;
extern int32 ms_type;
extern int32 mouse_mapping[3];
extern int32 accel[3];
extern int64 click_speed;
//extern sem_id kb_sem;
//extern sem_id ms_sem;
extern sem_id kbsem;
extern sem_id mssem;
extern char kbsem_name[];
extern char mssem_name[];
extern char kb_wait_sem_name[];
extern int8 ps2_mouse_found;
extern int32 typematic_rate;
extern int64 repeat_delay;
extern int32 ser_fd_count;
extern int32 *ser_fd_list;
extern int32 found_keyboard;
extern int64 control_alt_del_timeout;
extern int8 reset_pending;
extern int8 sample_rate;
extern int32 ps2_mouse_id;
extern int32 valid_rates[];
extern int32 valid_delays[];
extern int8 extended_key_75;

/* extern device_hooks kbm_device; */
extern int32 api_version;

/* BSS */

uint8 buff_46[3];
uint8 buffer_49[3];
uint8 k_locked_66[12];
isa_module_info *isa;
uint8 cmd_byte;
uint8 keyboardResult[16];
uint8 key_states[32];
uint8 non_repeating[32];
timer reset_timer;
char *kb_wmouse_devices[4];

#endif  /* KB_WMOUSE_H */
